import crypt
import functools
from collections import namedtuple

import requests
from cached_property import cached_property

from ocflib.infra import db

VHOST_MAIL_DB_PATH = '/home/s/st/staff/vhost/vhost-mail.conf'
VHOST_MAIL_DB_URL = 'https://www.ocf.berkeley.edu/~staff/vhost-mail.conf'

get_connection = functools.partial(db.get_connection, db='ocfmail')


class MailVirtualHost(namedtuple('MailVirtualHost', ('user', 'domain'))):

    def get_forwarding_addresses(self, c):
        """Return list of MailForwardingAddress objects."""
        c.execute(
            'SELECT `address`, `password`, `forward_to`, `last_updated`'
            'FROM `addresses`'
            'WHERE `address` LIKE %s',
            ('%@' + self.domain,),
        )
        return {
            MailForwardingAddress(
                address=r['address'],
                crypt_password=r['password'],
                forward_to=frozenset(addr.strip() for addr in r['forward_to'].split(',') if addr.strip()),
                last_updated=r['last_updated'],
            )
            for r in c
        }

    def add_forwarding_address(self, c, addr):
        # sanity check: forward_to should be a non-empty list-ish
        assert len(addr.forward_to) > 0, addr.forward_to
        assert not isinstance(addr.forward_to, str)

        c.execute(
            'INSERT INTO `addresses`'
            '(`address`, `password`, `forward_to`)'
            'VALUES (%s, %s, %s)',
            (addr.address, addr.crypt_password, ','.join(addr.forward_to)),
        )

    def remove_forwarding_address(self, c, addr):
        c.execute(
            'DELETE FROM `addresses`'
            'WHERE `address` = %s',
            (addr,),
        )


class MailForwardingAddress(namedtuple('MailForwardingAddress', (
    'address',
    'crypt_password',
    'forward_to',
    'last_updated',
))):

    @cached_property
    def is_wildcard(self):
        return self.address.startswith('@')


def get_mail_vhost_db():
    """Returns lines from the vhost database. Loaded from the filesystem (if
    available), or from the web if not."""
    try:
        with open(VHOST_MAIL_DB_PATH) as f:
            return list(map(str.strip, f))
    except IOError:
        # fallback to database loaded from web
        return requests.get(VHOST_MAIL_DB_URL).text.split('\n')


def get_mail_vhosts():
    """Returns a list of MailVirtualHost objects."""
    vhosts = set()
    for line in get_mail_vhost_db():
        line = line.strip()
        if not line or line.startswith('#'):
            continue
        user, domain = line.split()
        vhosts.add(MailVirtualHost(user=user, domain=domain))
    return vhosts


def vhosts_for_user(user):
    """Return MailVirtualHost objects for the user."""
    return {vhost for vhost in get_mail_vhosts() if vhost.user == user}


def crypt_password(password):
    """Return hashed password, compatible with the vhost database."""
    return crypt.crypt(password, salt=crypt.METHOD_SHA512)



"""Methods for figuring out how to identify the current user."""
import os
import pwd
from email.utils import formataddr

from ocflib.misc.mail import email_for_user


def current_uid():
    """Returns numerical UID of current user."""
    return os.getuid()


def current_user():
    """Returns username of current user."""
    return pwd.getpwuid(current_uid()).pw_name


def current_user_fullname():
    """Returns full name (from GECOS, not LDAP) of current user."""
    return pwd.getpwuid(current_uid()).pw_gecos


def current_user_email():
    """Returns @ocf email address of the current user."""
    return email_for_user(current_user())


def current_user_formatted_email():
    """Returns @ocf email address of the current user, including name.

    >>> current_user_formatted_email()
    'Chris Kuehl <ckuehl@ocf.berkeley.edu>'
    """
    return formataddr((current_user_fullname(), current_user_email()))



import os
import shlex
import string
import subprocess

import pexpect

KADMIN_PATH = '/usr/bin/kadmin'


def create_kerberos_principal_with_keytab(
    principal,
    keytab,
    admin_principal,
    password=None,
):
    """Creates a Kerberos principal by shelling out to kadmin.

    :param principal: name of the principal to create
    :param keytab: path to the admin keytab
    :param admin_principal: admin principal to authenticate with keytab
    :param password: password of the new principal (optional);
                     if not given, defaults to using a random password
    :return: the password of the newly-created account
    """
    # try changing using kadmin pexpect
    cmd = ('{kadmin} -K {keytab} -p {admin} add --use-defaults ' +
           '{principal}').format(
        kadmin=shlex.quote(KADMIN_PATH),
        keytab=shlex.quote(keytab),
        admin=shlex.quote(admin_principal),
        principal=shlex.quote(principal),
    )

    if not password:
        # XXX: using `--random-password` generates weak passwords, plus spits
        # them to stdout, so we just generate a random one ourselves
        allowed = string.ascii_letters + string.digits + string.punctuation
        password = ''.join(allowed[byte % len(allowed)]
                           for byte in os.urandom(100))

    child = pexpect.spawn(cmd, timeout=10)

    child.expect("{}@OCF.BERKELEY.EDU's Password:".format(principal))
    child.sendline(password)
    child.expect("Verify password - {}@OCF.BERKELEY.EDU's Password:"
                 .format(principal))
    child.sendline(password)

    child.expect(pexpect.EOF)

    output = child.before.decode('utf8')
    child.close()
    if child.exitstatus:
        raise ValueError('kadmin error: {}'.format(output))

    return password


def get_kerberos_principal_with_keytab(principal, keytab, admin_principal):
    """Returns information about an existing kerberos principal.

    Currently, this requires shelling out to kadmin, so the only information
    returned is whether the principal exists.

    :param principal: name of the principal to create
    :param admin_principal: name of the admin principal
    :param keytab: path to the admin keytab
    :return: True if the principal exists, or
             None if the principal does not exist
    """
    cmd = [
        shlex.quote(KADMIN_PATH),
        '-K', shlex.quote(keytab),
        '-p', shlex.quote(admin_principal),
        'get', shlex.quote(principal),
    ]

    try:
        subprocess.check_output(cmd, timeout=10, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        if b'Principal does not exist' in e.output:
            return None
        else:
            raise ValueError('kadmin error: {}'.format(e.output))

    return True
